/* This is source for a "base" for any normal PCI driver.
 * (But NOT a "STREAMS" type driver. There's a separate sample source
 * for that.
 * Both types are available from http://www.bolthole.com/solaris/drivers/
 *
 *
 * Read the README file for important information
 *    source version @(#) base.c 1.15@(#)
 * 
 * You can use this source for any purpose you like [without any warantee, etc]
 * It would be nice if you give me some credit if you use it, but you dont
 * have to.
 *
 * You should read the stuff at docs.sun.com about what to do if
 * you are writing a "block-transfer" based driver, as opposed to a
 * char-based driver.
 *
 * Either way, you'll have to provide routines for the appropriate entry
 * points in the cb_ops struct below
 * 
 * REMEMBER: If a C function isn't mentioned in
 *  "man -s9f intro", you pretty much cannot use it here.
 *
 * You'll want to globally replace "base_" with "yourdrv_",
 * as well as "base:" with "yourdrv:"
 *
 * You will also want to customize the 'struct base_softstate'
 * struct. In particular, if you want to map more than one register set
 * into kernel memory.
 * You will have to add the appropriate register mapping calls to
 * base_attach() to match what you put there.
 *
 * You'll also probably want to fiddle with the small dev_addr struct
 *
 *  DISCLAIMER: This code may not be 100% clean, kernel-wise.
 *  But it is as clean as I know how to make it, with lots of
 *  hopefully helpful comments.
 *    Philip Brown
 */

#include <stdarg.h>
#define _SYS_VARARGS_H		/* sys/varargs.h hack for gcc */

#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/stat.h>

/* THESE MUST BE LAST */
#include <sys/ddi.h>
#include <sys/sunddi.h>



/* This ASSUMES you create minor nodes where the higher bits denote
 * the instance of the driver. (aka "which card is this?")
 * The lower bits denote clone number, or anything else you want.
 */

#define DEVT_TO_INST(devt) (getminor(devt)>>4)
#define DEVT_TO_CLONENUM(devt) (getminor(devt) & 0xf)

/* Used to choose a node  num for a device node compatible with DEVT_TO_INST*/
#define INST_TO_NODENUM(instance) (instance<<4)

/* As well as changing the NAME of this, you will also want to change
 * the TYPE of the minor node created.
 */
#define DEVNODE_NAME	"base"	
#define DEVNODE_TYPE	DDI_NT_BLOCK /* see sunddi.h */

/*Our globally visible stuffs, that we populate the global
 * driver entry structs with  [base_cb_ops, base_ops]
 * Yeah, its kind of wierd to have a "static" be also externally visible,
 * but that's traditionally how these are declared.
 */
static	int	base_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
		    void **result);
static	int	base_probe(dev_info_t *dip);
static	int	base_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static	int	base_open(dev_t *dev, int openflags, int otyp, cred_t *credp);

static	int	base_print(dev_t, char*);

static	int	base_read(dev_t dev, struct uio* uiop, cred_t * credp);

static	int	base_close(dev_t dev, int openflags, int otyp, cred_t *credp);
static	int	base_write(dev_t dev, struct uio *uiop, cred_t *credp);
static	int	base_ioctl(dev_t dev, int cmd, intptr_t arg, int flags,
			   cred_t *credp, int *rvalp);
static	int	base_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);


static int debug_level;

/* "char/block operations" */
static struct cb_ops	base_cb_ops = {
	base_open,      /* open() */
	base_close,     /* close() */
	nodev,          /* strategy() */
	base_print,     /* print routine */
	nodev,          /* no dump routine */
	base_read,      /* read() */
	base_write,     /* write() */
	base_ioctl,     /* generic ioctl */
	nodev,          /* no devmap routine */
	nodev,          /* no mmap routine */
	nodev,          /* no segmap routine */
	nochpoll,       /* no chpoll routine */
	ddi_prop_op,
	0,              /* not a STREAMS driver	*/
	D_NEW | D_MP,   /* safe for multi-thread/multi-processor */
	CB_REV,         /* cb_ops version? */
	nodev,          /* cb_aread() */
	nodev,          /* cb_awrite() */
};

/* "device operations". */
static struct dev_ops base_ops = {
	DEVO_REV,            /* devo_rev */
	0,                   /* devo_refcnt */
	base_getinfo,        /* devo_getinfo */
	nulldev,             /* devo_identify */
	base_probe,          /* devo_probe */
	base_attach,         /* devo_attach */
	base_detach,         /* devo_detach */
	nodev,               /* devo_reset */
	&base_cb_ops,        /* devo_cb_ops */
	(struct bus_ops *)0, /* devo_bus_ops */
	NULL,                /* devo_power */
};

/* Ignore this. This is only a filler for the next struct, and
 * is already declared in <sys/modctl.h>
 */
extern	struct	mod_ops mod_driverops;

static	struct modldrv modldrv = {
	&mod_driverops,
	"Experimental _base_ driver",
	&base_ops,
};

static	struct modlinkage modlinkage = {
	MODREV_1,		/* MODREV_1 is indicated by manual */
	{&modldrv,NULL,NULL,NULL}

};


/************************************************************
 * And here's finally the vars specific to THIS driver,
 * that get allocated on a per-instance basis.
 * Here's where you should keep instance-specific mutexes,
 * DMA handles, register mapping handles, and stuff like that
 *
 ************************************************************/


/* This describes things like byteorder internally for the pci card */
static struct ddi_device_acc_attr dev_attr = {
	DDI_DEVICE_ATTR_V0,
	DDI_STRUCTURE_LE_ACC,
	DDI_STRICTORDER_ACC
};


#define MAX_INSTANCES 4 /* max of 4 hardware devices loaded */

/* 
 * Note that this mutex gets used in _init. Which means this mutex
 * isn't just for this "instance" of the driver. It is for ALL INSTANCES.
 * If you have multiple cards, you have multiple instances of the driver,
 * with this single shared mutex across them all
 */
static kmutex_t   base_global_mutex;

/* This is SORT OF like	int (*array)[] , except you NEVER USE THIS DIRECTLY.
 * You just use it with ddi_get_soft_state().
 * 
 *    "Global Softstate Handle" 
 */
void * base_glob_soft_handle;

/* This gets filled out on a per-instance basis by base_attach */
struct base_softstate {
	dev_info_t *dip; /* you'll probably want this saved! */
	kmutex_t instance_mutex;
	kmutex_t intr_mutex;
	ddi_iblock_cookie_t intr_cookie;
	
	ddi_acc_handle_t base_regs_handle; /*opaque kernel stuff to map regs*/
	struct base_regs *base_regs; /* actual pointer to mapped registerset*/
};


#ifdef YOUFILLTHISOUT
/* You fill out this struct to match the offsets in your device registers.
 * The "base_regs" pointer in base_softstate, above,  gets mapped
 * to your first set of device registers, in base_attach().
 *
 * BE CAREFUL!! WHen initially filling this out, you probably will want
 * to do a sizeof(struct base_regs) to see whether your compiler has
 * messed things up with padding
 *
 * Remove the ifdef after you've really filled this out.
 * Or just use byte offsets, and re-type struct base_regs to void*
 * in the softstate struct.
 */
struct base_regs {
	uint16_t somereg;
	uint32_t longerreg;
	
};
#endif

/*****************************************************************/
/*****************************************************************/
/*              Start of actual kernel routines here             */
/*****************************************************************/
/*****************************************************************/


void
base_debugprint(const char *format, ...)
{
	va_list args;
	va_start(args, format);

	if(debug_level==0){
		return;
	}

	/* DO NOT USE vcmn_err here!!! it will not be interrupt-safe.
	 * See the "CONTEXT" part of the manpage
	 */

	vprintf(format, args);
	va_end(args);
}


/*Like the base_global_mutex above, this routine is GLOBAL across all
 * instances.
 * It actually gets called when you add_drv your driver from the command
 * line, before  * it is hooked up to any specific hardware.
 * See base_attach() below for instance-specific stuff.
 *
 * HOWEVER... Its often nice to pre-allocate "softstate" space
 * here, if you dont mind hardcoding the total number of possible
 * instances you can have. So we do that.
 * 
 */
int
_init(void)
{
       int     ret;

	base_debugprint("base: _init called\n");
	/*
	 * Initialize global mutex before mod_install'ing driver.
	 * If mod_install() fails, must clean up mutex initialization
	 */
	mutex_init(&base_global_mutex, "base: Global Mutex",
                     MUTEX_DRIVER, (void *)NULL);

	/* This is kind of like
	 * malloc(sizeof(struct base_softstate *MAX_INSTANCES))
	 */
	ret=ddi_soft_state_init(&base_glob_soft_handle,
				sizeof(struct base_softstate),
			        MAX_INSTANCES);
	if(ret!=0){
		return (ret);
	}
	
	if ((ret = mod_install(&modlinkage)) != 0) {
		mutex_destroy(&base_global_mutex);
		ddi_soft_state_fini(&base_glob_soft_handle);
		cmn_err(CE_NOTE, "base: _init: mod_install FAILED:\n");
		return DDI_FAILURE;
	}

	cmn_err(CE_NOTE, "base: _init returning SUCCESS, supposedly\n");
	return (DDI_SUCCESS);
}

int
_info(struct modinfo *modinfop)
{
	
	cmn_err(CE_NOTE, "base: _info called\n");
	return (mod_info(&modlinkage, modinfop));
}

/* This is the match to the _init() routine. Clean up stuff allocated
 * specifically in _init() here.
 * Stuff in attach() should be cleaned up in detach()
 */
int
_fini(void)
{
	int ret;

	cmn_err(CE_NOTE, "base: _fini called\n");
	
        /*
         * If mod_remove() is successful, we destroy our global mutex
         */
	if ((ret = mod_remove(&modlinkage)) == 0) {
		ddi_soft_state_fini(&base_glob_soft_handle);
		mutex_destroy(&base_global_mutex);
	}
	return (ret);
}

/************************************************************/

/*
 * interrupt handler for the base device.
 * This wont ever get called until ddi_add_intr has been called.
 * Dont forget that you can only use intr_mutex in here, NOT
 * instance_mutex.
 * Additionally, you can ONLY USE FUNCTIONS that are
 * INTERRUPT CONTEXT SAFE.
 *  "user", or "kernel" context, just wont cut it.
 *
 * Note that cmn_err() is NOT interrupt-context safe.
 */
static uint_t
base_intr(caddr_t intr_arg)
{
	/* struct base_softstate *softstate;
	 * softstate = (struct base_softstate *)intr_arg;
	 */

	/* Put whatever you want here, IF you have interrupts */
	/* dont forget to call ddi_add_intr() to enable it! */

	return DDI_INTR_CLAIMED;
}

/* Kinda wierd, but this is an official solaris entry point, that can
 * be registered by cb_ops. We dont use it, but we provide it
 * for the OS
 */
static int base_print(dev_t dev, char*string){
	int instance=DEVT_TO_INST(dev);
	cmn_err(CE_NOTE, "base-print(%d): %s",instance,string);
	return 0;
}

static int base_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd,
			     void *arg, void **result){
	base_debugprint("base: getinfo called\n");
	return DDI_FAILURE;
}




/* you probably wont even need to have a probe for a PCI driver 
 * But just in case you want one to tell you what major and minor
 * you're being probed with, here's one
 */
 
static int base_probe(dev_info_t *dip){
#ifndef DO_PROBE_ROUTINE
   	base_debugprint("base: probe called\n");
#else

  	target=ddi_prop_get_int(DDI_DEV_T_ANY,dip,DDI_PROP_DONTPASS,
				"target", -1);
  	lun=ddi_prop_get_int(DDI_DEV_T_ANY,dip,DDI_PROP_DONTPASS,
				"lun", -1);
   
   	cmn_err(CE_NOTE, "probe- called fortarget=%d, lun=%d\n", target,lun);

	/* Normally, we would check the dev_info info,
	 * to see which device ID is being probed, and then
	 * check to see if the durn thing is actually there.
	 * But for now, assume .conf file is correct
	 */
   
	/* our two options for success are:
	 * DDI_PROBE_DONTCARE
	 * DDI_PROBE_SUCCESS
	 */

#endif /* DO_PROBE_ROUTINE */

	return DDI_PROBE_DONTCARE;
	/*return DDI_PROBE_FAILURE;*/
}

/* This routine can be called by base_attach (if errors occur)
 * and base_detach (on normal detach)
 * Idea is to clean up any mutexes, register mappings, etc. nicely.
 * We also free the instance-softstate.
 * We do NOT remove the minor node, however. Only detach does
 * that currently.
 */
static void base_cleanup(struct base_softstate *softstate){
	int instance;

	if(softstate==NULL){
		/* yes, paranoia is good*/
		cmn_err(CE_WARN,"ERROR: base_cleanup called with NULL");
		return;
	}
   	instance=ddi_get_instance(softstate->dip);

	if(softstate->intr_cookie != NULL){
		ddi_remove_intr(softstate->dip,0,softstate->intr_cookie);
	}
	if(softstate->base_regs_handle !=NULL){
		ddi_regs_map_free(&softstate->base_regs_handle);
		ddi_soft_state_free(base_glob_soft_handle, instance);
	}
	ddi_soft_state_free(base_glob_soft_handle,instance);
	
}


/*
 * _init() has already been called. This routine gets called
 * once for EVERY MATCHING PIECE OF HARDWARE.
 * It gets called for initial "attach", and also for resume(/suspend)
 * You would usually allocate space for state variables, and
 * map device registers in here.
 *
 * To avoid reboot loops, be very very careful putting things in here.
 * If you want to put extra debug type stuff in, it's safer putting
 * it in base_open() or someplace like that, that gets called after
 * the driver is safely loaded
 *
 * Be sure to free/deallocate anything you create here, in
 *  base_detach()
 */
static int base_attach(dev_info_t *dip, ddi_attach_cmd_t cmd){
   	int instance, status;
	struct base_softstate *softstate;
	int tmpint;

	/* default debug level to **1**
	 * You can override in the .conf file with "debug-level=0".
	 * When you have a production-ready driver, change
	 * default to 0
	 */
	debug_level=ddi_prop_get_int(DDI_DEV_T_ANY,dip, DDI_PROP_DONTPASS,
                                    "debug-level", 1);

	base_debugprint("base: base_attach called\n");

   	switch(cmd){
	 case DDI_RESUME:
	   	cmn_err(CE_NOTE, "base: Ignoring attach_RESUME\n");
		return DDI_FAILURE;
	 case DDI_PM_RESUME:
		cmn_err(CE_NOTE, "base: Ignoring attach_PM_RESUME\n");
		return DDI_FAILURE;
	 case DDI_ATTACH:
	   	break;
	 default:
	   	cmn_err(CE_NOTE, "base: unknown attach command\n");
	   	return DDI_FAILURE;
	}
   	instance=ddi_get_instance(dip);
   	base_debugprint("base: attach for instance %d\n", instance);


	/* We dont have to check for being over MAX_INSTANCES, because
	 * these two calls do it implicitly
	 */
	if (ddi_soft_state_zalloc(base_glob_soft_handle, instance) != DDI_SUCCESS) {
		cmn_err(CE_WARN, "base: %d: Failed to alloc soft state",
		    instance);
		return (DDI_FAILURE);
	}
	softstate=ddi_get_soft_state(base_glob_soft_handle,instance);
	if(softstate==NULL){
		cmn_err(CE_WARN, "base: attach for instance %d: Failed to get soft state",
			instance);
		ddi_soft_state_free(base_glob_soft_handle,instance);
		return (DDI_FAILURE);
	}

	softstate->dip = dip;

	/* you will probably want this at some point or another */
	mutex_init(&softstate->instance_mutex, NULL, MUTEX_DRIVER, NULL);



	if(ddi_dev_nregs(dip, &tmpint)==DDI_FAILURE){
		tmpint=0;
	}
	/* IFF there are registers to map, map the first set */
	if(tmpint>0){
		base_debugprint("base: %d register sets exist. Mapping #1\n",
				tmpint);

		status = ddi_regs_map_setup(dip, 1,
				(caddr_t *)&softstate->base_regs,
				0, 0,
				&dev_attr, &softstate->base_regs_handle);

		if (status != DDI_SUCCESS) {
			if (status == DDI_REGS_ACC_CONFLICT) {
				cmn_err(CE_WARN,
				 "regs mapping conflicted with other drivers");
				base_cleanup(softstate);
				return DDI_FAILURE;
			} else {
				cmn_err(CE_NOTE,
				 "regs mapping failed: no registers to map ");
			}
			
		}
	} else {
		base_debugprint("base: Note: there are no registers to map\n");
	}


	/* This will create a device node in the appropriate place
	 * under /devices/pci[*]/XXXXXX
	 *
	 * To get an automatic matching link in /dev/XXXX, you need to 
	 * add your own line to /etc/devlink.tab
	 */
	 
	ddi_create_minor_node(dip, DEVNODE_NAME, S_IFCHR,
			  INST_TO_NODENUM(instance),
			  DEVNODE_TYPE, 0);

	/* We want to do interrupt registering late in the game.
	 * Make sure everything is set up fully,
	 * just in case a random interrupt triggers the handler before
	 * we've even fully finished attach() !
	 */
	if(ddi_dev_nintrs(dip, &tmpint) == DDI_FAILURE){
		tmpint=0;
	}
	if(tmpint>0){
		base_debugprint("base: interrupt available. Registering.\n");
		/* And if you have interrupts, you should use THIS mutex for
		 * anything that the interrpt handler waits on
		 */
		ddi_get_iblock_cookie(dip, 0, &softstate->intr_cookie);
		mutex_init(&softstate->intr_mutex, NULL,
			   MUTEX_DRIVER, (void*)softstate->intr_cookie);

		if (ddi_add_intr(dip, 0,
		     &softstate->intr_cookie, NULL,
		     base_intr, (caddr_t)softstate) != DDI_SUCCESS) {

			cmn_err(CE_WARN, "Could not add interrupt handler");
			base_cleanup(softstate);
			ddi_remove_minor_node(dip, DEVNODE_NAME);
			return DDI_FAILURE;
		}
	}

	/* Dont forget to make ddi_remove_minor_node match in detach() */

	return DDI_SUCCESS;

	/* if you add anything to this routine, make ultra-paranoid
	 * checks on return values and return (DDI_FAILURE)
	 * if ANYTHING doesnt go completely 100% okay
	 */
}

static int base_open(dev_t *dev, int openflags, int otyp, cred_t *credp){
	int instance=DEVT_TO_INST(*dev);
	struct base_softstate *softstate;

   	base_debugprint("base: open called\n");

	/* even if you dont USE this, it is a check to see if we need
	 * to return ENXIO
	 */
	softstate=ddi_get_soft_state(base_glob_soft_handle, instance);
	if(softstate==NULL){
		/* This is technically NOT an error; it is expected.
		 * We return ENXIO to support "deferred attach"
		 */
		return (ENXIO);
	}

	/* return 0 on okay, or an errno otherwise */
   	return 0;
}



static int base_read(dev_t dev, struct uio* uiop, cred_t * credp){
	int instance=DEVT_TO_INST(dev);
	
   	base_debugprint("base: read called\n");
	/* If you are handling things directly here rather than through
	 * physio, make sure that uiop->resid is set to 0 on good read
	 */
	return (ENXIO);
}

static int base_write(dev_t dev, struct uio *uiop, cred_t *credp){
	int instance=DEVT_TO_INST(dev);

   	base_debugprint("base: write called\n");
	/* If you are handling things directly here rather than through
	 * physio, make sure that uiop->resid is set to 0 on good write
	 */
   	return (ENXIO);
}

static int base_close(dev_t dev, int openflags, int otyp, cred_t *credp){
   	base_debugprint("base: close called\n");
   	return (ENXIO);
}

static int base_ioctl(dev_t dev, int cmd, intptr_t arg, int flags,
                      cred_t *credp, int *rvalp){
   	base_debugprint("base: ioctl called\n");
   	return (ENXIO);
}

static int base_detach(dev_info_t *dip, ddi_detach_cmd_t cmd){
   	int instance;
	struct base_softstate *softstate;
   	instance=ddi_get_instance(dip);

	base_debugprint("base: detach for instance %d\n", instance);

	softstate=ddi_get_soft_state(base_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_NOTE,"base_detach failed: cannot get softstate?");
		return DDI_FAILURE;
	}

	base_cleanup(softstate);

	ddi_remove_minor_node(dip, DEVNODE_NAME);

	return (DDI_SUCCESS);
		     
}

